;; Generated package description from valign.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "valign" "3.1.1.0.20260416.235856" "Visually align tables" '((emacs "26.0")) :commit "c63faceba523f22f73df152cffcd54c41f997acd" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("convenience" "text" "table") :url "https://github.com/casouri/valign")
